  <div id="about-excerpt">
    {% assign article = pages.frontpage %}
    {% if article.content != "" %}
      <h2>{{ article.title }}</h2>
      {{ article.content }}
    {% else %}
      In <em>Admin &gt; Blogs &amp; Pages</em>, create a page with the handle <strong><code>frontpage</code></strong> and it will show up here.<br />
      {{ "Learn more about handles" | link_to: "http://wiki.shopify.com/Handle" }}
    {% endif %}
  </div>

  <table id="gallery">
  {% tablerow product in collections.frontpage.products cols: 3 limit: 12 %}
    <div class="gallery-image">
      <a href="{{ product.url | within: collections.frontpage }}" title="{{ product.title | escape }} &mdash; {{ product.description | strip_html | truncate: 50 | escape }}"><img src="{{ product.images.first | product_img_url: 'small' }}" alt="{{ product.title | escape }}" /></a>
    </div>
    <div class="gallery-info">
      <a href="{{ product.url | within: collections.frontpage }}">{{ product.title | truncate: 30 }}</a><br />
      <small>{{ product.price | money }}{% if product.compare_at_price_max > product.price %} <del>{{ product.compare_at_price_max | money }}</del>{% endif %}</small>
    </div>
  {% endtablerow %}
  </table>
